# PUPPy, A Playstation Update Package (PUP) packer/unpacker
# Copyright (C) 2011, Hexodin

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, struct, hmac, hashlib

class PUP:
    def __init__(self, hmac_key_path):
        """Class to pack and unpack Playstation 3 Update Packages."""
        self.hmac_key = open(hmac_key_path, "rb").read()
        self.file_list = {0x100 : "version.txt",
                          0x101 : "license.txt",
                          0x102 : "promo_flags.txt",
                          0x103 : "update_flags.txt",
                          0x104 : "patch_build.txt",
                          0x200 : "ps3swu.self",
                          0x201 : "vsh.tar",
                          0x202 : "dots.txt",
                          0x203 : "patch_data.pkg",
                          0x300 : "update_files.tar"}

    def unpack(self, pup_filename, dir_name):
        """Unpack Playstation 3 Update Packages."""
        try: pup_file = open(pup_filename, "rb")
        except IOError, error:
            raw_input(error)
            sys.exit()
        if pup_file.read(8)[0:5] != "SCEUF":
            raw_input("This is not a SCE Update File")
            sys.exit()
        pup_file.seek(16, 1)
        file_count = struct.unpack(">Q", pup_file.read(8))[0]
        pup_file.seek(16, 1)
        file_info = [[struct.unpack(">Q", pup_file.read(8))[0] for x in range(4)] for y in range(file_count)]
        if not os.path.exists(dir_name): os.mkdir(dir_name)
        for data in file_info:
            if self.file_list.has_key(data[0]):
                actual_file = open(os.path.join(dir_name, self.file_list[data[0]]), "wb")
                pup_file.seek(data[1], 0)
                actual_file_data = pup_file.read(data[2])
                actual_file.write(actual_file_data)
        pup_file.close()

    def pack(self, new_pup_filename, dir_name, pkg_ver = 1, img_ver = 0xb99c):
        """Pack Playstation 3 Update Packages."""
        if not os.path.exists(dir_name):
            raw_input("Directory not found.")
            sys.exit()
        try: new_pup_file = open(new_pup_filename, "w+b")
        except IOError, error:
            raw_input(error)
            sys.exit()
        file_list = os.listdir(dir_name)
        num_files = len(file_list)
        table_len = num_files * 32
        header_len = 0x50 + (2 * table_len)
        new_pup_file.write("SCEUF\x00\x00\x00")
        new_pup_file.write(struct.pack(">Q", pkg_ver))
        new_pup_file.write(struct.pack(">Q", img_ver))
        new_pup_file.write(struct.pack(">Q", num_files))
        new_pup_file.write(struct.pack(">Q", header_len))
        overall_size = 0
        file_info_table = list()
        hash_table = list()
        file_data = list()
        count = 0
        for data in self.file_list.keys():
            if self.file_list[data] in file_list:
                actual_file = open(os.path.join(dir_name, self.file_list[data]), "rb")
                actual_file_size = os.path.getsize(actual_file.name)
                actual_file_data = actual_file.read()                
                actual_file_hash = hmac.new(self.hmac_key, actual_file_data, hashlib.sha1)
                file_info_table.append(data)
                if count == 0: file_info_table.append(header_len)
                else: file_info_table.append(file_info_table[len(file_info_table) - 4] + file_info_table[len(file_info_table) - 3])
                file_info_table.append(actual_file_size)
                file_info_table.append(0)
                hash_table.append(actual_file_hash)
                file_data.append(actual_file_data)
                overall_size += actual_file_size
                count += 1
                actual_file.close()
        new_pup_file.write(struct.pack(">Q", overall_size))
        for file_info_data in file_info_table:
            new_pup_file.write(struct.pack(">Q", file_info_data))
        for num in range(len(hash_table)):
            new_pup_file.write(struct.pack(">Q", num))
            new_pup_file.write(hash_table[num].digest())
            new_pup_file.write("\x00" * 4)
        new_pup_file.seek(0, 0)
        pup_header = new_pup_file.read(header_len)
        header_hash = hmac.new(self.hmac_key, pup_header, hashlib.sha1)
        new_pup_file.close()
        the_pup_file = open(new_pup_filename, "a+b")
        the_pup_file.write(header_hash.digest())
        the_pup_file.write("\x00" * 0xc)
        for files in file_data: the_pup_file.write(files)
        the_pup_file.close()
